#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>

// #define __USE_XOPEN  // For strptime
#include <time.h>

#include <math.h>
#define pi 3.14159265358979323846


/* 
Description for a track point
*/

struct track_point {
    double lat, lon;
    int time;
};

double deg2rad(double deg) {
  return (deg * pi / 180);
}

double rad2deg(double rad) {
  return (rad * 180 / pi);
}

/*
  credit to: https://www.geodatasource.com/developers/c
  for the distance function
*/
double distance(double lat1, double lon1, double lat2, double lon2, char unit) {
  double theta, dist;
  if ((lat1 == lat2) && (lon1 == lon2)) {
    return 0;
  }
  else {
    theta = lon1 - lon2;
    dist = sin(deg2rad(lat1)) * sin(deg2rad(lat2)) + cos(deg2rad(lat1)) * cos(deg2rad(lat2)) * cos(deg2rad(theta));
    dist = acos(dist);
    dist = rad2deg(dist);
    dist = dist * 60 * 1.1515;
    switch(unit) {
      case 'M':
        break;
      case 'K':
        dist = dist * 1.609344;
        break;
      case 'N':
        dist = dist * 0.8684;
        break;
    }
    return (dist);
  }
}

int intFloor(double d){
  return (int)d;
}

double squareRoot(double d){
  return sqrt(d);
}

double doubleFloor(double d){
  return floor(d);
}

// https://stackoverflow.com/questions/26376396/library-function-to-extract-substring-by-position-index
char * substring(char *c, int startint, int endint){
  char *start = &c[startint];
  char *end = &c[endint];
  // Note the + 1 here, to have a null terminated substring
  char *substr = (char *)calloc(1, end - start + 1);
  memcpy(substr, start, end - start);
  return substr;
}

// https://www.epochconverter.com/programming/c
int convert_time(char *timestr){
    // printf("time in convert     %s\n", timestr);
    int year = strtol(substring(timestr, 0, 4), NULL, 0);
    int month = strtol(substring(timestr, 5, 7), NULL, 0);
    int mday = strtol(substring(timestr, 8, 10), NULL, 0);
    int hour = strtol(substring(timestr, 11, 13), NULL, 0);
    int min = strtol(substring(timestr, 14, 16), NULL, 0);
    int sec = strtol(substring(timestr, 17, 19), NULL, 0);
    struct tm t;
    time_t t_of_day;

    t.tm_year = year-1900;  // Year - 1900
    t.tm_mon = month;           // Month, where 0 = jan
    t.tm_mday = mday;          // Day of the month
    t.tm_hour = hour;
    t.tm_min = min;
    t.tm_sec = sec;
    t.tm_isdst = -1;        // Is DST on? 1 = yes, 0 = no, -1 = unknown
    t_of_day = mktime(&t);

    // printf("seconds since the Epoch: %ld\n", (long) t_of_day);
    return t_of_day;
}

xmlXPathObjectPtr getnodeset (xmlDocPtr doc, xmlChar *xpath, xmlXPathContextPtr context){
    xmlXPathObjectPtr result;    

    if (context == NULL) {
        printf("Error in xmlXPathNewContext\n");
        return NULL;
    }
    result = xmlXPathEvalExpression(xpath, context);
    if (result == NULL) {
        printf("Error in xmlXPathEvalExpression\n");
        xmlXPathFreeObject(result);
        return NULL;
    }
    if(xmlXPathNodeSetIsEmpty(result->nodesetval)){
        xmlXPathFreeObject(result);
        printf("No result\n");
        return NULL;
    }
    return result;
}

void print_tps(struct track_point *tps, int tps_size) {
    printf("(");
    for (int j=0; j < tps_size; j++) {
        printf("(%f, %f, %d),", tps[j].lat, tps[j].lon, tps[j].time);
    } 
    printf(")\n");
}
int sizeTp(struct track_point *tps) {
   int s = sizeof(tps);
   return s;
}

void printFirst(struct track_point *tps){
   printf("%f, %f, %d \n", tps[0].lat, tps[0].lon, tps[0].time);
}

double totDist(struct track_point *tps) {
    //returns total distance in km
    int tps_size = tps[0].time; //retrieve size
    double tot_dist = 0;
    //printf("%ld\n", sizeof(tps));
    for (int j=1; j < tps_size; j++) {
        double tmp = distance(tps[j].lat, tps[j].lon, tps[j+1].lat, tps[j+1].lon, 'K');
        tot_dist+=tmp;
    } 
    return tot_dist;
}
int totTime(struct track_point *tps) {
    //returns total time in mins
    int tps_size = tps[0].time;
    int tot_time = 0;
    int startTime = tps[1].time;
    int endTime = tps[tps_size].time;
    tot_time = endTime-startTime;

    return tot_time/60;
}

double avgSpeed(struct track_point *tps){
    //returns avg. speed in km/h
    double dist = totDist(tps);
    double time = (double)totTime(tps); //in hours
    time = time/60.0;
    //printf("%f\n", time);
    return dist/time;
}

void stat(struct track_point *tps){
    printf("******** GPX FILE SUMMARY ********\n");
    printf("Distance Covered: %f km\n", totDist(tps));
    printf("Elapsed Time: %d min\n", totTime(tps));
    printf("Average Speed: %f km/h\n", avgSpeed(tps));
    printf("**********************************\n");
}

struct track_point * parseGPX(char *filepath) {
    // Load file
    xmlDocPtr doc = xmlParseFile(filepath);

    if (doc == NULL) {
        fprintf(stderr, "Document not parsed successfully.");
    }
    xmlNodePtr cur = xmlDocGetRootElement(doc);

    if (cur == NULL) {
        fprintf(stderr, "empty document");
        xmlFreeDoc(doc);
        return 0;
    }

    xmlChar *xpath = (xmlChar *) "//gpx:trk/gpx:trkseg/gpx:trkpt";
    xmlChar *timexpath = (xmlChar*) "//gpx:trk/gpx:trkseg/gpx:trkpt/gpx:time";
    xmlXPathContextPtr context = xmlXPathNewContext(doc);
    xmlXPathRegisterNs(context, (xmlChar*)"gpx", (xmlChar*)"http://www.topografix.com/GPX/1/1");

    xmlNodeSetPtr nodeset;
    xmlNodeSetPtr timenodeset;
    xmlXPathObjectPtr result;
    xmlXPathObjectPtr timeset;
    int i = 0;
    //xmlChar *keyword;

    struct track_point *tps;

    result = getnodeset(doc, xpath, context);
    timeset = getnodeset(doc, timexpath, context); 
    xmlXPathFreeContext(context);

    if (result) {
        nodeset = result->nodesetval;
        int size = nodeset->nodeNr;

        // printf("size %d", size);
        tps = (struct track_point *)malloc(sizeof(struct track_point)*(size)+2);

        //stash away size
        struct track_point s;
        s.lon = s.lat = 0.0;
        s.time = size;
        tps[0] = s;

        timenodeset = timeset->nodesetval;
        for (i=0; i < nodeset->nodeNr; i++) {

            xmlNodePtr ptNode = result->nodesetval->nodeTab[i];

            struct track_point tp;

            xmlChar *lat = xmlGetProp(ptNode, (xmlChar*)"lat");
            if (lat != 0) {
                //printf("lat %s", lat);
                tp.lat = strtof((char *)lat, NULL);
                //fix
                //printf("tp.lat %f", tp.lat);
                xmlFree(lat);
            }
            xmlChar *lon = xmlGetProp(ptNode, (xmlChar*)"lon");
            if (lat != 0) {
                //printf("lon %s ", lon);
                tp.lon = strtof((char *)lon, NULL);
                xmlFree(lon);
            }
            xmlChar *time = xmlNodeListGetString(doc, timenodeset->nodeTab[i]->xmlChildrenNode, 1);
            if (time != 0) {
                // printf("time %s ", time);
                int time_store;
                time_store = convert_time((char *) time);
                //printf("converted time %d\n", time_store);
                tp.time = time_store;
                xmlFree(time);
            }           
            tps[i+1] = tp;
            //printf(" testing %f %f %d\n", tps[i].lat, tps[i].lon, tps[i].time);
        }  
        xmlXPathFreeObject(timeset);
        xmlXPathFreeObject(result);
        //printf("Printing array of track_points: \n");
        //print_tps(tps, size);
        xmlFreeDoc(doc);
        xmlCleanupParser();
        return tps;
    }
    else {
        return NULL;
    }
    
}

int numPt(char *filepath) {
    xmlDocPtr doc = xmlParseFile(filepath);

    if (doc == NULL) {
        fprintf(stderr, "Document not parsed successfully.");
    }
    xmlNodePtr cur = xmlDocGetRootElement(doc);

    if (cur == NULL) {
        fprintf(stderr, "empty document");
        xmlFreeDoc(doc);
        return 0;
    }

    xmlChar *xpath = (xmlChar *) "//gpx:trk/gpx:trkseg/gpx:trkpt";
    xmlXPathContextPtr context = xmlXPathNewContext(doc);
    xmlXPathRegisterNs(context, (xmlChar*)"gpx", (xmlChar*)"http://www.topografix.com/GPX/1/1");

    xmlNodeSetPtr nodeset;
    xmlXPathObjectPtr result;

    result = getnodeset(doc, xpath, context);
    xmlXPathFreeContext(context);
    int size = 0;

    if (result) {
        nodeset = result->nodesetval;
        size = nodeset->nodeNr;
    }
    xmlXPathFreeObject(result);
    xmlFreeDoc(doc);
    xmlCleanupParser();
    return size; 
}



#ifdef BUILDTEST
int main(int argc, char *argv[]){
    if (argc < 2) {
        printf("Wrong number of arguments");
    }
    printf("Number of track points: %d\n", numPt(argv[1]));
    struct track_point * tps = parseGPX(argv[1]);
    double dist = totDist(tps);
    printf("size: %d\n", tps[0].time);
    printf("total distance: %f\n", dist);
    int time = totTime(tps);
    printf("total time: %d\n", time);
    double speed = avgSpeed(tps);
    printf("avg speed: %f\n", speed);




    //free(tps);
}
#endif
