(*
   OCamlLex scanner for GWiz
   File: scanner.mll
   Authors: Ashley Kim, Katherine Duff
   PLT Spring 2021

*)

{ open Gwizparse }

(* Definitions *)
let digit = ['0' - '9']
let letter = ['a'-'z' 'A'-'Z']
let digits = digit+
let simple_char = [' '-'!' '#'-'&' '('-'[' ']'-'~']
let escape_char = ['t' 'r' 'n' '\'' '\"' '\\']

(* Rules *)
rule token = parse
  [' ' '\t' '\r'] {token lexbuf}
| "/*"  { comment lexbuf }
| '\n'  { Lexing.new_line lexbuf; token lexbuf }

(* Blocks, Delimiters, Terminators *)
| '('   {LPAREN}
| ')'   {RPAREN}
| '{'   {LBRACE}
| '}'   {RBRACE}
| '['   {LBRAC}
| ']'   {RBRAC}
| ';'   {SEMI}
| '.'   {PERIOD}
| ','   {COMMA}

(* Arithmetic Operators *)
| '+'   {PLUS}
| '-'   {MINUS}
| '*'   {TIMES}
| '/'   {DIVIDE}

(* Assignment Operators *)
| '='   {ASSIGN}

(* Relational Operators *)
| "=="  {EQ}
| "!="  {NEQ}
| "<"   {LT}
| "<="  {LEQ}
| '>'   {GT}
| ">="  {GEQ}

(* Logical Operators *)
| "&&"  {AND}
| "||"  {OR}
| '!'   {NOT}
| '-'   {NEG}

(* Keywords *)
| "if"          {IF}
| "else"        {ELSE}
| "for"         {FOR}
| "while"       {WHILE}
| "return"      {RETURN}
| "int"         {INT}
| "double"      {DOUBLE}
| "bool"        {BOOL}
| "string"      {STRING}
| "void"        {VOID}
| "Trackpoint"  {TRACKPOINT}
| "file"        {FILE}

(* Literals *)
| "true"                                {BLIT(true)}
| "false"                               {BLIT(false)}
| digits as lit                         {LITERAL(int_of_string lit)}
| digits '.' digit* ( ['e' 'E'] ['+' '-']? digits )? as lit {DLIT(lit)}
| letter (letter | digit)* as lit       {ID(lit)}
| '\"' ((simple_char | '\\' escape_char)* as lit) '\"'   {SLIT(lit)}
| eof   {EOF}

and comment = parse
    "*/"        { token lexbuf }
| _             { comment lexbuf }
